@@ -8,6 +8,8 @@ import openai
 import scipy.stats
 import transformers
 import shap
+
+import adatest
 from ._model import Model
 from .utils import isinstance_ipython
 #from transformers.tokenization_utils_base import ExplicitEnum
@@ -77,6 +79,241 @@ class ClassifierScorer(Scorer):
     that input.
     """
 
+    def __init__(self, model, topk=1, output_names=None, method="dirichlet", dirichlet_concentration=10):
+        """ Create a new scorer given a model that returns a probability vector for each input string.
+        
+        Parameters:
+        -----------
+        model : callable
+            A model that is callable with a single argument (which is a list of strings) and returns a matrix of outputs.
+
+        topk : int
+            The number of top outputs to consider when scoring tests. For example topk=2 causes "should not be" tests to
+            check the top two model outputs.
+
+        output_names : list of strings
+            A list of strings that correspond to the outputs of the model. If None, model.output_names is used.
+
+        method : 'margin' or 'dirichlet'
+            The scoring method to use. Dirichlet is preferred, but margin is available for backwards compatibility.
+
+        dirichlet_concentration : float
+            The concentration parameter for the dirichlet scoring method. It is in the units o pseudo-counts where larger
+            values lead to a tighter prior centered around the model's probability outputs (so scores are more likely to
+            be -1 or +1).
+        """
+        super().__init__(model)
+
+        # we don't want to re-init a class if init has alrady been done (this can happen when Scorer(maybe_scorer) is called)
+        if hasattr(self, "output_names"):
+            return # already initialized
+
+        # extract output names from the model if they are not provided directly
+        if output_names is None:
+            self.output_names = self.model.output_names
+        else:
+            self.output_names = output_names
+        
+        if not callable(self.output_names):
+            self._output_name_to_index = {v: i for i, v in enumerate(self.output_names)}
+            assert topk == 1, "topk must be 1 right now" # TODO: this is because we need to figure out topk and templates
+        self.topk = topk
+        self.method = method
+        self.dirichlet_concentration = dirichlet_concentration
+
+    def __call__(self, tests, score_column, overwrite_outputs=False, output_bias=0.5):
+        """ Score a set of tests.
+
+        Parameters
+        ----------
+        tests : pandas.DataFrame
+            A dataframe of tests.
+
+        output_bias : float
+            How much to bias the output entries in a test towards the actual output from the model.
+        """
+
+
+        
+        # determine which rows we need to evaluate
+        eval_inputs = []
+        eval_inds = []
+        for i, (id, test) in enumerate(tests.iterrows()):
+            if test[score_column] == "" and test.label != "topic_marker":
+                template_expansions = expand_template(test.input)
+                for expansion in template_expansions:
+                    eval_inputs.append(expansion)
+                    eval_inds.append(i)
+
+        # run the model on the rows we need to evaluate
+        try:
+            model_out = self.model(eval_inputs)
+        except Exception as e:
+            model_out = np.zeros((len(eval_inputs), len(self.model.output_names))) * np.nan # TODO: remove this hack after the user study
+            log.error(e)
+            log.error(eval_inputs)
+            log.error("The model threw an exception when evaluating inputs! We are patching this disaster with np.nan for the sake of the user study!")
+
+        # compute the output strings for each output
+        out_strings = [[] for _ in range(tests.shape[0])]
+        out_probs = [[] for _ in range(tests.shape[0])]
+        i = 0
+        while i < len(model_out):
+            out_strings[eval_inds[i]].append(self.model.output_names[np.argmax(model_out[i])])
+            out_probs[eval_inds[i]].append(np.max(model_out[i]))
+            i += 1
+        for i in eval_inds:
+            out_strings[i] = "|".join(out_strings[i]) # template outputs are joined by |
+            out_probs[i] = np.min(out_probs[i]) # the probability of a set of items is the prob of the min item
+
+        # ensure the model output is represented in the tests
+        current_outputs = tests["output"]
+        current_labelers = tests["labeler"]
+        updated_ids = []
+        for i, ind in enumerate(eval_inds):
+            if current_labelers.iloc[ind] == "imputed":
+                id = tests.index[ind]
+                tests.loc[id, "output"] = out_strings[ind]
+                tests.loc[id, "label"] = ""
+                tests.loc[id, score_column] = out_probs[ind]
+                updated_ids.append(id)
+            elif not overwrite_outputs and current_outputs.iloc[ind] != out_strings[ind]:
+
+                # mark the current row as nan score (meaning the output does not match)
+                tests.loc[tests.index[ind], score_column] = np.nan
+
+                # add a new test where the model output does match
+                id = uuid.uuid4().hex
+                tests.loc[id, "topic"] = tests.loc[tests.index[ind], "topic"]
+                tests.loc[id, "input"] = eval_inputs[i]
+                tests.loc[id, "output"] = out_strings[ind]
+                tests.loc[id, "label"] = ""
+                tests.loc[id, "labeler"] = "imputed"
+                tests.loc[id, score_column] = out_probs[ind]
+
+                updated_ids.append(id)
+            else:
+                id = tests.index[ind]
+                tests.loc[id, "output"] = out_strings[ind]
+                tests.loc[id, score_column] = out_probs[ind]
+                updated_ids.append(id)
+        tests.deduplicate() # make sure any duplicates we may have introduced are removed
+
+        # reimpute missing labels
+        tests.impute_labels() # TODO: ensure this method caches the local models and only reimputes when needed for each topic
+
+        # set the test score sign to match the label
+        for id in updated_ids:
+            if id in tests.index and tests.loc[id, "label"] == "pass":
+                tests.loc[id, score_column] = -float(tests.loc[id, score_column])
+        
+
+
+        # out_pos = 0
+        # i = 0
+        # top_outputs = [{} for _ in range(tests.shape[0])]
+        # value2_outputs = [{} for _ in range(tests.shape[0])]
+        # while i < len(model_out):
+        #     out_pos = eval_inds[i]
+
+        #     # save the top model outputs
+        #     inds = np.argsort(-model_out[i])
+        #     shown_tmp = {}
+        #     for j in inds[:5]:
+        #         shown_tmp[self.model.output_names[j]] = float(model_out[i][j])
+        #     top_outputs[out_pos] = shown_tmp
+
+        #     out[out_pos].append(self.model.output_names[np.argmax(model_out[i])])
+
+        #     if output_string == tests.iloc[out_pos]['output']:
+        #         out[out_pos].append(float(model_out[i][top_ind]))
+
+        #     # To make sharing test trees between models slightly easier we fall back to looking for different
+        #     # capitalizations of the output if the original one doesn't exist
+        #     token_to_check = tests.iloc[out_pos]['output']
+        #     if token_to_check not in self._output_name_to_index:
+        #         if token_to_check.capitalize() in self._output_name_to_index:
+        #             token_to_check = token_to_check.capitalize()
+        #         elif token_to_check.lower() in self._output_name_to_index:
+        #             token_to_check = token_to_check.lower()
+            
+        #     # multiple tokens can be checked at the same time with templates
+        #     out_val = np.nan
+        #     ind = self._output_name_to_index.get(token_part, None)
+        #     if ind is not None and model_out[i] is not None:
+        #         sorted_values = np.argsort(model_out[i])
+        #         topk = topk_threshold_ind(ind, sorted_values, self.topk)
+
+        #         if self.method == "dirichlet":
+        #             raw_score = compute_dirichlet_score(
+        #                 ind, model_out[i], self.topk,
+        #                 concentration=self.dirichlet_concentration,
+        #                 # we treat values less than 10% of the topk value as unlikely to impact the results
+        #                 # this is used to avoid unnecessary computation
+        #                 domination_threshold=model_out[i][topk] / 10 
+        #             )
+
+        #             if test_type == "{} should output {}":
+        #                 score = raw_score
+        #             else:
+        #                 score = -raw_score
+
+        #         if np.isnan(model_out[i][ind]):
+        #             score = np.nan
+        #         elif model_out[i][ind] > model_out[i][topk]:
+        #             if self.method == "dirichlet":
+        #                 # mval = 1 / len(model_out[i]) if self.topk == 1 else 0 # minimum value possible while being at the top
+        #                 # score = (raw_score - mval) / (1 - mval) # scale from 0 to 1
+        #                 score = raw_score
+        #             else:
+        #                 score = model_out[i][ind] - model_out[i][topk]
+        #         else:
+        #             if self.method == "dirichlet":
+        #                 # mval = 1 / (self.topk + 1) # maximum value possible while not being at the top
+        #                 # score = (raw_score - 1 + mval) / (1 - mval) # scale from 0 to 1
+        #                 score = raw_score - 1
+        #             else:
+        #                 mask = (model_out[i] <= model_out[i][topk]) & (model_out[i] > model_out[i][ind])
+        #                 score = (model_out[i][ind] - model_out[i][mask]).sum()
+        #         if test_type == "{} should output {}":
+        #             score *= -1
+        #         # out_val = max(score, out_val)
+        #         out[out_pos].append(score)
+        #     # out[out_pos] = max(out[out_pos], out_val)
+        #     i += 1
+
+
+        #     # out_pos += 1
+        # return {
+        #     "scores": out,
+        #     "outputs": outputs,
+        #     "value2_outputs": value2_outputs
+        # }
+
+    def suggest_outputs(self, current, num_suggestions=20):
+        prompt = ""
+        for c in current:
+            prompt += '"'+c+'"\n'
+        prompt += '"{output}'
+        response = openai.Completion.create(
+            engine='curie-instruct-beta', prompt=[prompt.format(output=o) for o in self.output_names], max_tokens=0, # self.engine
+            temperature=0, n=1, stop='\"', logprobs=0, echo=True
+        )
+        lines = [sum(choice["logprobs"]["token_logprobs"][11:]) for choice in response["choices"]]
+        pairs = list([v for v in zip(lines, self.output_names) if v[1] not in current])
+        pairs.sort()
+        return [v[1] for v in list(reversed(pairs))[:num_suggestions]]
+
+
+class ClassifierScorerOld(Scorer):
+    """ Wraps a model and defines a callable scorer that returns a score value for any input/output pair.
+
+    Positive scores indicate test failures, positive scores indicate tests that pass. For example if we wrap
+    a text sentiment classifer the `scorer(TestTree([("this is great!", "should be", "POSITIVE")]))` will return
+    a large positive value indicating that the model is very likely to correctly produce that output when given
+    that input.
+    """
+
     def __init__(self, model, topk=1, output_names=None, method="dirichlet", dirichlet_concentration=10):
         """ Create a new scorer given a model that returns a probability vector for each input string.
         
@@ -314,12 +551,11 @@ class ClassifierScorer(Scorer):
         pairs.sort()
         return [v[1] for v in list(reversed(pairs))[:num_suggestions]]
 
-
 class GeneratorScorer(Scorer):
     """ Wraps a text generation model in a scorer that can score the target model against tests.
     """
 
-    def __init__(self, model, reverse_model=None, embedding_model=None, feature_models={}, similarity_threshold=0.9):
+    def __init__(self, model, completions=1, reverse_model=None, feature_models={}, similarity_threshold=0.9):
         """ Initializes a new scorer for a given target model.
 
         Parameters
@@ -328,15 +564,15 @@ class GeneratorScorer(Scorer):
             The model we are scoring against the tests. It is expected to be a function that takes a list of strings as
             input and returns a list of strings as output.
 
+        completions : int
+            The number of completions to generate for each model input.
+
         reverse_model : callable
             The inverse of model, such that x = reverse_model(model(x)). Note that this is optional and only required
             to use the 'should be invertable' test type. If for example model is an EN to ES translation model, then
             reverse_model should be an ES to EN translation model. Just like model, reverse_model is expected to take a
             list of strings as input and return a list of strings as output.
 
-        embedding_model : callable
-            Used to embed the tests for comparison purposes. TODO: shouldn't this be shared with the main adatest backend?
-
         feature_models : dict
             A dictionary of classifiers that can be used to score the generated output. The keys are the names of the
             classifier label that is predicted and the values are the classifiers themselves. Each classifer is expected
@@ -354,25 +590,28 @@ class GeneratorScorer(Scorer):
         if hasattr(self, "supported_test_types"):
             return # already initialized
 
+        self.completions = completions
         self.reverse_model = reverse_model
-        self.embedding_model = embedding_model
         self.similarity_threshold = similarity_threshold
         self.feature_models = feature_models
 
         self.supported_test_types = [
-            "{} should output {}",
-            "{} should not output {}", # TODO: should this use semantic similarity instead of exact string match?
-            "{} should have the same output as {}",
-            "{} should output text containing {}",
-            "{} should not output text containing {}",
-            "{} should output text that is {}",
-            "{} should not output text that is {}",
-            "{} should not be completed to become {}"
+            "{} can output {}",
+            "{} can only output {}",
+            "{} cannot output {}", # TODO: should this use semantic similarity instead of exact string match?
+            "{} can have the same output as {}",
+            "{} can output text containing {}",
+            "{} cannot output text containing {}",
+            "{} can output text starting with {}",
+            "{} cannot output text starting with {}",
+            "{} can output text that is {}",
+            "{} cannot output text that is {}",
+            "{} cannot be completed to become {}"
         ]
         # '"{}[]" should not become "{}" when completed. 
 
         # see if we can support the inversion test
-        if self.reverse_model is not None and self.embedding_model is not None: # TODO: do we need an embeddig model or just use the backend?
+        if self.reverse_model is not None and adatest.embedding_model is not None: # TODO: do we need an embeddig model or just use the backend?
             self.supported_test_types.append("{} should be invertable. [] is the output.") # Note that {} means user-editable, [] means read-only
 
         # # see if we have user-provided classifier tests
@@ -397,7 +636,8 @@ class GeneratorScorer(Scorer):
         eval_io_feature_pos = {k: [] for k in self.feature_models}
         for i, (k, test) in enumerate(tests.iterrows()):
             if test.type in ["{} should not output {}", "{} should output {}",
-                             "{} should not output text containing {}", "{} should output text containing {}"]:
+                             "{} should not output text containing {}", "{} should output text containing {}",
+                             "{} should not output text starting with {}", "{} should output text starting with {}"]:
                 v1 = expand_template(test.value1)
                 for s1 in v1:
                     eval_inputs.append(s1)
@@ -444,14 +684,14 @@ class GeneratorScorer(Scorer):
                         eval_inds.append(i)
                 variations1.append(v1)
                 variations2.append(v2)
-        try:
-            # model_out = self.model(eval_inputs)
-            model_out = self.model(eval_inputs, completions=self.completions) # now a list of lists
-        except Exception as e:
-            model_out = ["ERROR" for _ in range(len(eval_inputs))]#np.zeros((len(eval_inputs), len(self.model.output_names))) * np.nan # TODO: remove this hack after the user study
-            log.error(e)
-            log.error(eval_inputs)
-            log.error("The model threw an exception when evaluating inputs! We are patching this disaster with 'ERROR' for the sake of the user study!")
+        # try:
+            # model_out = self.model(eval_inputs) {} should {} {}
+        model_out = self.model(eval_inputs, completions=self.completions) # now a list of lists
+        # except Exception as e:
+        #     model_out = [["__ERROR__"]*self.completions for _ in range(len(eval_inputs))]#np.zeros((len(eval_inputs), len(self.model.output_names))) * np.nan # TODO: remove this hack after the user study
+        #     log.error(e)
+        #     log.error(eval_inputs)
+        #     log.error("The model threw an exception when evaluating inputs! We are patching this disaster with '__ERROR__'!")
 
         # run feature models when we need to
         model_output_feature = {}
@@ -459,7 +699,8 @@ class GeneratorScorer(Scorer):
         model_io_feature = {}
         for k in self.feature_models:
             if len(eval_output_feature_pos[k]) > 0:
-                feature_out = self.feature_models[k]([model_out[ind] for ind in eval_output_feature_pos[k]])
+                feature_out_flat = self.feature_models[k]([v for ind in eval_output_feature_pos[k] for v in model_out[ind]])
+                feature_out = np.reshape(feature_out_flat, (-1, 2))
                 model_output_feature[k] = [None for _ in model_out]
                 for i, ind in enumerate(eval_output_feature_pos[k]):
                     model_output_feature[k][ind] = feature_out[i]
@@ -469,7 +710,8 @@ class GeneratorScorer(Scorer):
                 for i, ind in enumerate(eval_input_feature_pos[k]):
                     model_input_feature[k][ind] = feature_out[i]
             if len(eval_io_feature_pos[k]) > 0:
-                feature_out = self.feature_models[k]([eval_inputs[ind] + model_out[ind] for ind in eval_io_feature_pos[k]])
+                feature_out_flat = self.feature_models[k]([eval_inputs[ind] + v for ind in eval_io_feature_pos[k] for v in model_out[ind]])
+                feature_out = np.reshape(feature_out_flat, (-1, self.completions))
                 model_io_feature[k] = [None for _ in model_out]
                 for i, ind in enumerate(eval_io_feature_pos[k]):
                     model_io_feature[k][ind] = feature_out[i]
@@ -497,7 +739,7 @@ class GeneratorScorer(Scorer):
             round_trip_embed = [None for _ in model_out]
             try:
                 # compute input embedding
-                tmp = self.embedding_model.encode([eval_inputs[ind] for ind in eval_reverse_pos], convert_to_tensor=True, show_progress_bar=False).cpu()
+                tmp = adatest.embedding_model.encode([eval_inputs[ind] for ind in eval_reverse_pos], convert_to_tensor=True, show_progress_bar=False).cpu()
                 for i, ind in enumerate(eval_reverse_pos):
                     input_embed[ind] = tmp[i]
 
@@ -507,7 +749,7 @@ class GeneratorScorer(Scorer):
                     model_reverse_out[ind] = str(reverse_out[i])
 
                 # compute round trip embedding
-                tmp = self.embedding_model.encode(reverse_out, convert_to_tensor=True, show_progress_bar=False).cpu()
+                tmp = adatest.embedding_model.encode(reverse_out, convert_to_tensor=True, show_progress_bar=False).cpu()
                 for i, ind in enumerate(eval_reverse_pos):
                     round_trip_embed[ind] = tmp[i]
 
@@ -527,24 +769,30 @@ class GeneratorScorer(Scorer):
             out_pos = eval_inds[i]
 
             test_type = tests.iloc[out_pos]["type"]
-            if test_type == "{} should not output {}" or test_type == "{} should output {}" or test_type == "{} should not output text containing {}":
-                invert = -1 if test_type == "{} should not output {}" or test_type == "{} should not output text containing {}" else 1
+            if test_type in ["{} should not output {}", "{} should output {}",
+                             "{} should output text containing {}", "{} should not output text containing {}",
+                             "{} should output text starting with {}", "{} should not output text starting with {}"]:
+                invert = -1 if test_type in ["{} should not output {}", "{} should not output text containing {}", "{} should not output text starting with {}"] else 1
                 contain_check = test_type == "{} should not output text containing {}" or test_type == "{} should output text containing {}"
+                starting_with_check = test_type == "{} should not output text starting with {}" or test_type == "{} should output text starting with {}"
                 
-                # auto fill missing outputs
+                # auto fill missing outputs TODO: this is a hack
                 if tests.iloc[out_pos]['value2'] is None:
-                    tests.loc[tests.index[out_pos], 'value2'] = str(model_out[i])
+                    tests.loc[tests.index[out_pos], 'value2'] = str(model_out[i][0])
                 
                 # save the model output
-                value1_outputs[out_pos]  = {}
-                value1_outputs[out_pos][model_out[i]] = 1
+                value1_outputs[out_pos] = {}
+                value1_outputs[out_pos]["string"] = list(model_out[i])
 
                 # multiple tokens can be checked at the same time with templates
                 for token_part in expand_template(tests.iloc[out_pos]['value2']):
-                    if contain_check:
-                        out[out_pos].append(-invert if token_part in model_out[i] else invert)
-                    else:
-                        out[out_pos].append(-invert if model_out[i] == token_part else invert)
+                    for c in range(self.completions):
+                        if contain_check:
+                            out[out_pos].append(-invert if token_part in model_out[i][c] else invert)
+                        elif starting_with_check:
+                            out[out_pos].append(-invert if token_part.startswith(model_out[i][c]) else invert)
+                        else:
+                            out[out_pos].append(-invert if model_out[i][c] == token_part else invert)
                 i += 1
             elif test_type == "{} should output text that is {}" or test_type == "{} should not output text that is {}":
                 should_be = test_type == "{} should output text that is {}"
@@ -570,29 +818,32 @@ class GeneratorScorer(Scorer):
             elif test_type == "{} should not be completed to become {}":
 
                 # multiple features can be checked at the same time with templates
+                io_feature_probs = []
+                input_feature_probs = []
                 for feature in expand_template(tests.iloc[out_pos]['value2']):
-                    if feature in model_io_feature:# and feature in model_input_feature is implied
-                        io_feature_prob = model_io_feature[feature][i]
-                        input_feature_prob = model_input_feature[feature][i]
-                        if input_feature_prob <= 0.5 and io_feature_prob >= 0.5:
-                            score = io_feature_prob - input_feature_prob
-                        elif input_feature_prob <= 0.5 and io_feature_prob <= 0.5:
-                            score = io_feature_prob - input_feature_prob - 0.5
-                        elif input_feature_prob >= 0.5 and io_feature_prob >= 0.5:
-                            score = io_feature_prob - input_feature_prob + 0.5
-                        else: # if input_feature_prob >= 0.5 and io_feature_prob <= 0.5
-                            score = io_feature_prob - input_feature_prob
-                    else:
-                        score = np.nan
-                    out[out_pos].append(score)
+                    for c in range(self.completions):
+                        if feature in model_io_feature:# and feature in model_input_feature is implied
+                            io_feature_prob = model_io_feature[feature][i][c]
+                            input_feature_prob = model_input_feature[feature][i]
+                            if input_feature_prob <= 0.5 and io_feature_prob <= 0.5 or input_feature_prob >= 0.5 and io_feature_prob >= 0.5:
+                                score = io_feature_prob - input_feature_prob - 0.5
+                            else: # if input_feature_prob >= 0.5 and io_feature_prob <= 0.5 or (input_feature_prob <= 0.5 and io_feature_prob >= 0.5)
+                                score = io_feature_prob - input_feature_prob
+                        else:
+                            input_feature_prob = io_feature_prob = 0
+                            score = np.nan
+                        out[out_pos].append(score)
+                        input_feature_probs.append(input_feature_prob)
+                        io_feature_probs.append(io_feature_prob)
                 
                 # save the model output
                 value1_outputs[out_pos] = {}
-                value1_outputs[out_pos]["string"] = model_out[i]
+                value1_outputs[out_pos]["string"] = list(model_out[i])
 
+                max_score_ind = np.argmax(out[out_pos])
                 value2_outputs[out_pos] = {}
-                value2_outputs[out_pos]["input prob"] = input_feature_prob
-                value2_outputs[out_pos]["input+output prob"] = io_feature_prob
+                value2_outputs[out_pos]["input prob"] = input_feature_probs[max_score_ind]
+                value2_outputs[out_pos]["input+output prob"] = io_feature_probs[max_score_ind]
 
                 i += 1
             
@@ -632,7 +883,7 @@ class GeneratorScorer(Scorer):
         }
 
 
-def expand_template(s):
+def expand_template(s, keep_braces=False):
     """ Expand a template string into a list of strings.
     """
     # parts = []
@@ -641,7 +892,10 @@ def expand_template(s):
     s = re.sub("{[^}]*}", "{}", s)
     template_groups = [str(m)[1:-1].split("|") for m in matches]
     try:
-        return [s.format(*parts) for parts in itertools.product(*template_groups)]
+        if keep_braces:
+            return [s.format(*['{{{p}}}' for p in parts]) for parts in itertools.product(*template_groups)]
+        else:
+            return [s.format(*parts) for parts in itertools.product(*template_groups)]
     except ValueError:
         return [s] # we return the template not filled in if it is invalid
 
